----------------
-- File globals
----------------
local shortName = "SomeTracker"
local Addon = LibStub("AceAddon-3.0"):NewAddon(shortName, "AceConsole-3.0", "AceEvent-3.0")
Addon.name = "Some Tracker"
Addon.shortName = shortName
Addon.settings = LibStub:GetLibrary(Addon.shortName .. "-Settings")
Addon.version = 0.1

local D = LibStub("DoomCore-0.6")
local K = D.K
local assertType = D.assertType
Addon.handler  = D.Handler:New(Addon)

local playerID, playerName
local inWorld
local unique = D.NodeCrawler:New()
local templateCache = D.NodeCrawler:New()

local frames = {}
local spacing = 5
local fadeDuration = 0.6
--local missingIcon = "Interface\\Icons\\INV_Misc_QuestionMark"

--------
-- Core
--------

function Addon:OnInitialize()
  self.handler:Register{
    media = "LibSharedMedia-3.0",
    masque = "Masque",
    templates = "SomeTemplates-0.2"
  }

  self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
  self:RegisterEvent("PLAYER_ENTERING_WORLD")
  self:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED", "UpdateSpec")
  self:RegisterEvent("LEARNED_SPELL_IN_TAB", "UpdateSpec")
  self:RegisterEvent("BAG_UPDATE", "UpdateInv")
end

function Addon:OnLoad(registered)
  self:UpdateSpec()
  self:UpdateInv()

  for trackerName, tracker in self:GetAll(self.core, "tracker") do
    self:BuildTracker(trackerName)
  end
end

function Addon:Migrate()
  local oldVersion = self:GetVersion()
  if not oldVersion then
    self:MigrateDB({}, { "Default Damage" }, { "Default Healing" })
    self:Make(self.core, "tracker", "Default Damage")
    self:Make(self.core, "tracker", "Default Healing")
    return true
  end
end


function Addon:OnReset(registered)
  --if registered == false and self:Migrate() then return end
  local damage = self:Make(self.core, "tracker", "Default Damage")
  if not rawget(damage,"overflow") then damage.overflow = "addition" end
  local healing = self:Make(self.core, "tracker", "Default Healing")
  healing.trackerType = "healing"
  if not rawget(healing, "offsetX") then healing.offsetX = -damage.offsetX end
end


-------------------
-- Tracker editing
-------------------

function Addon:AddTracker(info, name)
  assertType(info,"table", name,"string")
  local tracker = self:Make(self.core, "tracker", name, true)
  if not tracker then return end
  self:BuildTracker(name)
end
function Addon:RenameTracker(info, new)
  assertType(info,"table", new,"string")
  if self.core[new] and self.core[new].type then return end
  local old = info[1]
  self.core[new] = self.core[old]
  self.core[old] = nil

  self.settings.crawler.old = nil
  frames[new[1]] = frames[old[1]]
  frames[old[1]] = nil
  self:BuildTracker(new)
end
function Addon:DeleteTracker(info)
  assertType(info,"table")
  local tracker = info[1]
  self:Delete(self.core, tracker)
  self.settings.options.args[tracker] = nil
end

function Addon:AddMeter(info, name)
  assertType(info,"table", name,"string")
  local trackerName = info[1]
  local tracker = self.core[trackerName]

  if #info < 4 then
    local meterType = tracker.add and tracker.add.meterType or "spell"
    tracker[meterType] = tracker[meterType] or {}
    local meter = self:Make(tracker[meterType], "meter", name, true)
    if not meter then return end
    self:BuildMeter(trackerName, meterType, name, meter)
    self:AddSubMeter(trackerName, meterType, name, meter, meterType, name)
  else
    local meter = tracker[info[2]][info[3]]
    self:AddSubMeter(trackerName, info[2], info[3], meter, meter.meterType, name)
  end
end
function Addon:DeleteMeter(info)
  assertType(info,"table")
  info = K(info)
  self:ConfDelete(info:r(1))
  self.settings.crawler:Set(K{ info[1], "args", info[2], "args", info[3] }, nil)
end

function Addon:AddSubMeter(trackerName, meterType, meterName, meter, subType, subName)
  assertType(trackerName,"string", meterType,"string", meterName,"string", meter,"table", subType,"string", subName,"string")
  subType = subType or "spell"
  if not self.settings.meterTypes[subType] then return end
  meter[subType] = meter[subType] or {}
  local subMeter = self:Make(meter[subType], "subMeter", subName, true)
  if not subMeter then return end
  meter.unmodified = nil
  local image
  if subType == "aura" then
    local name,_,image,_,_,_,_,_,_,id = UnitAura("player", subName)
    if subName == name then
      self:Make(subMeter, "alias", id, true)
    elseif subName == id then
      self:Make(subMeter, "alias", name, true)
    end
  elseif subType == "spell" then
    local spellInfo = { GetSpellInfo(subName) }
    image = spellInfo[3]
    local id = spellInfo[7]
    if id and tostring(subName) ~= tostring(id) then
      self:Make(subMeter, "alias", id, true)
    end
  end
  self.settings:BuildMeterSettings(trackerName, meterType, meterName, meter)
end
function Addon:DeleteSubMeter(info)
  assertType(info,"table")
  local packedKey = info[#info]
  local packedI = string.find(packedKey, "-")
  local subType = string.sub(packedKey, 1, packedI - 1)
  local name = string.sub(packedKey, packedI + 1)
  local trackerName, meterType, meterName = unpack(info)
  local meter = self:GetParent(info)
  meter.unmodified = nil
  self:Delete(meter[subType], name)
  self.settings:BuildMeterSettings(trackerName, meterType, meterName, meter)
end

function Addon:AddAlso(info, val)
  assertType(info,"table", val,{"string","number"})
  local trackerName, meterType, meterName = unpack(info)
  local meter = self:GetParent(info)
  local also = self:Make(meter, "also", val, true)
  if not also then return end
  meter.unmodified = nil
  also.image = select(3,GetSpellInfo(val))
  also.color = meter.newColor
  meter.newColor = meter.color
  self.settings:BuildMeterSettings(trackerName, meterType, meterName, meter)
end
function Addon:DeleteAlso(info)
  assertType(info,"table")
  local parent = self:Get(K(info):r(2))
  parent.unmodified = nil
  self:Delete(parent, info[#info-1])
end
function Addon:TrackerOverride(info, value)
  self:ConfSet(info, value)
  local trackerName = info[1]
  local overrideName = info[#info]

  for type,name, meter in self:GetAll(self.core[trackerName], "meter", 3) do
    meter[overrideName] = value
  end
end

local function alsoColor(info)
  return K(info):r(1):a(string.sub(info[#info], 0, -string.len("color")-1), "color")
end
function Addon:ConfGetAlsoColor(info)
  return self:ConfGet(alsoColor(info))
end
function Addon:ConfSetAlsoColor(info, r, g, b, a)
  return self:ConfSet(alsoColor(info), r, g, b, a)
end

function Addon:SetIcon(meter, image)
  meter.image = image
  return image
end

----------
-- Frames
----------

function Addon:UpdateFrame(info, value)
  self:ConfSet(info, value)
  self:BuildTrackerFrame(info[1])
end

function Addon:BuildTracker(trackerName)
  assertType(trackerName,"string")
  local tracker = self.core[trackerName]
  self.settings:BuildTrackerSettings(trackerName, tracker)
  self:BuildTrackerFrame(trackerName, tracker)
  for meterType,meterName, meter in self:GetAll(tracker, "meter", 3) do
    self:BuildMeter(trackerName, meterType, meterName, meter)
  end
end

function Addon:BuildTrackerFrame(name, tracker)
  if not tracker then tracker = self.core[name] end
  assertType(name,"string", tracker,"table")
  self:AutoDefault(tracker)
  local frame = frames[name]
  if not frame then
    frame = CreateFrame("Frame", self.shortName .. name, UIParent)
    frame.tex = frame:CreateTexture("ARTWORK");
    frame.tex:SetAllPoints();
    frame.tex:SetTexture(1, 1, 1); frame.tex:SetAlpha(0.5);

    frame:RegisterForDrag("LeftButton")
    frame:SetScript("OnDragStart", frame.StartMoving)
    frame:SetScript("OnDragStop", frame.StopMovingOrSizing)
    frame:SetScript("OnReceiveDrag", function()
      local x,y = select(4,frame:GetPoint(1))
      tracker.offsetX = math.floor(x*10)/10
      tracker.offsetY = math.floor(y*10)/10
      self.lib.registry:NotifyChange(self.name)
    end)
    frames[name] = frame
    if self.lib.masque then
      frame.msq = self.lib.masque:Group(self.name,name)
    end
  end
  frame.count = 0
  frame:ClearAllPoints()
  local size = tracker.iconSize
  frame:SetSize(size, size)
  frame:SetPoint("CENTER", nil, tracker.anchor, tracker.offsetX, tracker.offsetY)
  self:Clear(name)
  local locked = tracker.lock
  frame:SetMovable(not locked)
  frame:EnableMouse(not locked)
  if locked then
    frame.tex:Hide()
  else
    frame.tex:Show()
  end
  frame:Show()
  return frame
end

function Addon:BuildMeter(trackerName, meterType, meterName, meter)
  assertType(trackerName,"string", meterType,"string", meterName,"string", meter,"table")
  self.settings:BuildMeterSettings(trackerName, meterType, meterName, meter)
end

function Addon:GetFreeButton(trackerName, tracker, meterType)
  assertType(trackerName,"string", tracker,"table", meterType,"string")
  local frame = frames[trackerName]
  local button
  for _,b in ipairs{ frame:GetChildren() } do
    if not b:IsVisible() then
      button = b
      break
    end
  end
  if not button then
    frame.count = frame.count + 1
    button = CreateFrame("Button", frame:GetName() .. "-" .. frame.count, frame)
    button:EnableMouse(false)
    button:Disable()
    button.icon = button:CreateTexture(nil, "BACKGROUND")
    --button.icon:SetTexture(missingIcon)
    button.icon:SetAllPoints(button)
    button.text = button:CreateFontString(button:GetName() .. "-Amount")
    local animations = button:CreateAnimationGroup()
    local fade = animations:CreateAnimation("Alpha")
    fade:SetStartDelay(tracker.duration)
    fade:SetDuration(fadeDuration)
    fade:SetFromAlpha(1)
    fade:SetToAlpha(0)
    animations:SetScript("OnFinished", function()
      button:Hide()

      self:Sort(frame)
    end)
    button.animation = animations

    if frame.msq then frame.msq:AddButton(button, { Icon = button.icon }) end
  end

  local font = self.lib.media:Fetch(self.lib.media.MediaType.FONT, tracker.font)
  button.text:SetFont(font, tracker.fontSize, tracker.outline == "outline" and "OUTLINE")
  if tracker.outline == "shadow" then
    button.text:SetShadowColor(0,0,0,1)
    button.text:SetShadowOffset(1,-1)
  else
    button.text:SetShadowColor(0,0,0,0)
  end
  local size = tracker.iconSize
  button:SetSize(size, size)
  button.text:SetPoint("LEFT", size + 5, 1)
  button.meterType = meterType
  button:Show()
  return button
end

function Addon:Clear(trackerName)
  assertType(trackerName,"string")
  local children = { frames[trackerName]:GetChildren() }
  if #children == 0 then return end
  for _,child in ipairs(children) do
    self:RemoveAlert(trackerName, child)
  end
end

function Addon:RemoveAlert(trackerName, button)
  assertType(trackerName,"string", button,"table")
  local frame = frames[trackerName]
  button:Hide()
  button:ClearAllPoints()

  --button.icon:SetTexture(missingIcon)
  button.image = nil
  button.meterType = nil
  button.meter = nil
  button.text:SetText()
  button.text:SetTextColor(1,1,1,1)
end

local function sortFunc(e1, e2)
  local nil1 = e1 == nil or e1.time == nil
  local nil2 = e2 == nil or e2.time == nil
  if nil1 and nil2 then return false end
  if nil1 then return false end
  if nil2 then return true end
  return e2.time < e1.time
end

function Addon:Sort(frame)
  local children = { frame:GetChildren() }
  if #children == 0 then
    unique = D.NodeCrawler:New()
    return
  end
  local spells = {}
  local other = {}
  for _,child in ipairs(children) do
    if child:IsVisible() then
      if child.meterType == "spell" then table.insert(spells, child)
      else table.insert(other, child) end
    end
  end
  if #spells > 0 then table.sort(spells, sortFunc) end
  if #other > 0 then table.sort(other, sortFunc) end
  local add = {}
  for _,other in ipairs(other) do table.insert(add, other) end
  for _,spell in ipairs(spells) do table.insert(add, spell) end
  if #add == 0 then return end
  add[1]:ClearAllPoints()
  add[1]:SetPoint("TOPLEFT", frame)
  for i=2,#add do
    add[i]:ClearAllPoints()
    add[i]:SetPoint("TOPLEFT", add[i-1], "BOTTOMLEFT", 0, -spacing)
  end

  if frame.msq then frame.msq:ReSkin() end
end

function Addon:BuildAlert(trackerName, meterType, match, meterName, id, alertName, amount, crit, time, over)
  local tracker = self.core[trackerName]
  local displayOver = tracker.over
  if displayOver ~= "include" and displayOver ~= "additional" and amount == 0 then return end
  local frame = frames[trackerName]
  local alert
  for _,child in pairs{ frame:GetChildren() } do
    if child.meterType == meterType and child.alertName == alertName and child:IsVisible() then
      alert = child
      break
    end
  end
  if not alert then
    local meter = tracker[meterType][meterName]
    alert = self:GetFreeButton(trackerName, tracker, meterType)
    local image
    if meter.override then
      image = meter.image
    elseif meterType == "aura" then
      image = select(3, UnitAura("player", match)) or select(3, UnitAura("player", meterName))
    else
      image = select(3, GetSpellInfo(match)) or select(3, GetSpellInfo(meterName))
    end
    if not image and meterType == "spell" then image = select(3, GetSpellInfo(id)) end
    local oldImage = meter.image
    if image and not oldImage then meter.image = image end
    if not image and oldImage then image = oldImage end
    local color = meter[match] and meter[match].color
    if not color then color = meter.color end
    if color then alert.text:SetTextColor(color.r, color.g, color.b, color.a) end
    alert.icon:SetTexture(image)
    alert.displayCrit = meter.markCrits
    alert.amount = 0
    alert.over = 0
    alert.meterType = meterType
    alert.alertName = alertName
    alert.time = time
    alert.count = 0
    self:Sort(frame)
  end

  alert.animation:Stop()
  alert.animation:Play()

  alert.count = alert.count + 1
  alert.amount = alert.amount + amount

  local displayAmount = alert.amount

  if over then alert.over = alert.over + over end
  if displayOver == "include" then
    displayAmount = displayAmount + alert.over
  end

  local amountString = D.roundString(displayAmount)
  alert.crit = alert.displayCrit and (alert.crit or crit)
  if meterType == "spell" and alert.crit then
    amountString = amountString .. "*"
  end
  if alert.over >= 1000 and displayOver == "additional" then
    amountString = amountString .. " (" .. D.roundString(alert.over) .. ")"
  end
  if meterName == "Touch of Karma" then
    amountString = amountString .. "/" .. D.roundString(UnitHealthMax("player")/2)
  end
  alert.text:SetText(amountString)

  return alert
end


-------------------
-- Prebuilt meters
-------------------

function Addon:UpdateSpec()
  if not inWorld then return end
  self.settings:UpdateSpec()
  for meterName in pairs(self.settings.fromSpellbook) do
    self:AddDefault(meterName)
  end
  for meterName in pairs(self.settings.fromOther) do
    self:AddDefault(meterName)
  end
  --[[for meterName in pairs(self.settings.fromPvp) do
    self:AddDefaultSpell(meterName)
  end]]--
end
function Addon:UpdateInv()
  if not inWorld then return end
  self.settings:UpdateInv()

  for itemID in pairs(self.settings.fromItem) do
    itemID = tonumber(itemID)
    self:AddDefault(itemID)
  end
end

function Addon:TemplateKey(name)
  assertType(name,{"string","number"})
  local canonical
  local key
  local num =  tonumber(name)
  if num then
    canonical = self.lib.templates.canonical.items[num]
    if canonical then return { "items", canonical } end
  elseif self.lib.templates.all[name] then
    return { "all", name }
  else
    local class = select(2, UnitClass("player"))
    local spec = GetSpecialization()
    canonical = self.lib.templates.canonical[class][spec][name]
    if canonical then return { class, spec, canonical } end
  end
  canonical = self.lib.templates.canonical.items[name]
  if canonical then return { "items", canonical } end
end

function Addon:AddDefault(name)
  assertType(name,{"string","number"})
  local template = self:MeterFromTemplate(self:TemplateKey(name))
  if template.d then self:AddFromTemplate({"Default Damage", name}, false, true) end
  if template.h then self:AddFromTemplate({"Default Healing", name}, false, true) end
end

function Addon:AddFromTemplate(info, allowExcluded, fromDefault)
  assertType(info,"table")
  local trackerName = info[1]
  local name = info[#info]
  local key = self:TemplateKey(name)
  local template = self:MeterFromTemplate(key)
  if not template then return end
  if template.name and not template.name:find("%%") then name = template.name end
  local meterType = template.type or "spell"
  local tracker = self.core[trackerName]
  tracker[meterType] = tracker[meterType] or {}
  local trackerWithType = tracker[meterType]
  if allowExcluded == false and self:Excluded(trackerWithType, name) then
    return
  end
  local meter = trackerWithType[name]
  if meter and meter.type and not meter.unmodified then return end
  meter = self:Make(trackerWithType, "meter", name)
  self:DeployTemplate(meter, template)
  if fromDefault then meter.fromDefault = true end
  template._default = key
  meter.unmodified = true
  self:BuildMeter(trackerName, meterType, name, meter)
end

function Addon:DeployTemplate(meter, template)
  assertType(meter,"table", template,"table")
  D.setIndex(meter, template)
  for alsoName,color in pairs(template.also or {}) do
    if not self:Excluded(meter, alsoName) and not self:Overriden(meter, alsoName) then
      local also = self:Make(meter, "also", alsoName)
      also.color = D.hexToCol(color)
    end
  end
  for subType, i, subMeterName in D.npairs(template.meters, 3) do
    meter[subType] = meter[subType] or {}
    local meterWithType = meter[subType] or {}
    if not self:Excluded(meterWithType, subMeterName) and not self:Overriden(meterWithType, subMeterName) then
      self:Make(meterWithType, "subMeter", subMeterName)
    end end
end

function Addon:MeterFromTemplate(key)
  assertType(key,{"table","nil"})
  if not key then return end
  local cached = templateCache:Get(key)
  if cached then return cached end
  local template
  if #key == 2 then
    template = self.lib.templates[key[1]][key[2]]
  elseif #key == 3 then
    template = self.lib.templates[key[1]][key[2]][key[3]]
  end
  if not template then return end
  local name = key[#key]
  local meter = {}
  for k,v in pairs(template) do meter[k] = v end

  if type(name) == "number" then
    meter.image = GetItemIcon(name)
    meter.override = true
  elseif meter.override and meter.id then
    meter.image = select(3, GetSpellInfo(meter.id))
  end
  if meter.color then meter.color = D.hexToCol(meter.color) end
  for also,color in pairs(template.also or {}) do
    local icon = select(3, GetSpellInfo(also))
    if icon then
      meter.image = meter.image or icon
      meter.color = meter.color or D.hexToCol(color)
    end
  end
  if not meter.image then
    meter.image = select(3, GetSpellInfo(meter.name)) or (select(3,GetSpellInfo(name))) or (select(3,GetSpellInfo(template.id)))
  end
  meter.newColor = meter.color
  D.setIndex(meter, self.settings.defaults.meter)
  meter._default = key
  templateCache:Set(key, meter)
  return meter
end


----------
-- Events
----------

function Addon:PLAYER_ENTERING_WORLD()
  playerID = UnitGUID("player")
  playerName = UnitName("player")
  if not inWorld then
    self:CrawlDefaults()
    for tracker, meterType, meterName, meter in self:GetAll(self.core, "meter", 4) do
      local key = K{ tracker, meterType, meterName }
      local delete = false
      local templateKey = meter._default
      local template = self:MeterFromTemplate(templateKey)
      if templateKey and not template then
        delete = true
        templateKey = nil
      end

      if not templateKey and (meter.unmodified or meter.fromDefault) then
        templateKey = self:TemplateKey(meterName)
        template = self:MeterFromTemplate(templateKey)
        if template then
          meter._default = templateKey
        else
          delete = true
        end
      end
      if templateKey then
        delete = true
        if template then
          self:DeployTemplate(meter, template)

          local templateType = template.type or "spell"
          if templateType ~= meterType and meter.unmodified then
            delete = true
            self.core[tracker][templateType][meterName] = meter
            self.settings.crawler:Set({ tracker, "args", templateType, "args", meterName }, nil)
            self.settings:BuildMeterSettings(tracker, templateType, meterName, meter)
          elseif meter.fromDefault and meter.unmodified then
            delete = true
          else
            delete = false
          end
        end
      end
      if delete then
        self.settings.crawler:Set({ tracker, "args", meterType, "args", meterName }, nil)
        self.core[tracker][meterType][meterName] = nil
      end
      meter.newColor = meter.color
    end
  end

  inWorld = true
  self:UpdateSpec()
  self:UpdateInv()
end

function Addon:Match(subType, subName, subMeter, sourceName, petMatch, spellName, spellID)
  if subType == "spell" then
    local match = false
    if subName == spellID then
      self:Make(subMeter, "alias", spellName)
      return true
    elseif subName == spellName then
      self:Make(subMeter, "alias", spellID)
      return true
    end
  elseif petMatch and (subType == "pet" and subName == sourceName or
      (subName == "%n" and sourceName == playerName)) then return true
  elseif subType == "aura" then
    local name,_,icon,_,_,_,_,_,_,_,id = UnitAura("player", subName)
    if subName == name then
      self:Make(subMeter, "alias", id)
      return true
    elseif subName == id then
      self:Make(subMeter, "alias", name)
      return true
    end
  end
end

function Addon:Matches(meterType, meterName, meter, sourceName, petMatch, spellName, spellID)
  if self:Match(meterType, meterName, meter, sourceName, petMatch, spellName, spellID) then return meterName, spellID end

  for alsoName, also in self:GetAll(meter, "also") do
    if self:Match(meterType, alsoName, also, sourceName, petMatch, spellName, spellID) then return alsoName, spellID end
  end
  for subType,subName, subMeter in self:GetAll(meter, "subMeter", 3) do
    if self:Match(subType, subName, subMeter, sourceName, petMatch, spellName, spellID) then return meterName, spellID end
  end
end

function Addon:COMBAT_LOG_EVENT_UNFILTERED(
event,time,suffix,hide,
sourceGUID,sourceName,sourceFlags,sourceRaidFlags,
destGUID,destName,destFlags,destRaidFlags,
spellID,spellName,school,amount, overkill, school, resisted, blocked, absorbed, crit
)
  local kind
  if suffix == "SPELL_CAST_SUCCESS" then
    kind = "spell"
  elseif suffix == "SWING_DAMAGE" then
    amount = spellID
    spellID = nil
    spellName = nil
    kind = "damage"
  elseif suffix == "SPELL_DAMAGE" or suffix == "SPELL_BUILDING_DAMAGE"
      or suffix == "SPELL_PERIODIC_DAMAGE" or suffix == "RANGE_DAMAGE" then
    kind = "damage"
  elseif suffix == "SPELL_HEAL" or suffix == "SPELL_BUILDING_HEAL"
      or suffix == "SPELL_PERIODIC_HEAL" or suffix == "RANGE_HEAL" then
    kind = "healing"
  elseif suffix == "SPELL_AURA_APPLIED" or suffix == "SPELL_AURA_REFRESH" or suffix == "SPELL_AURA_REMOVED" then
    kind = "aura"
  else return end
  local unitMatch = sourceGUID == playerID
  local petMatch = not unitMatch and (
  CombatLog_Object_IsA(sourceFlags, COMBATLOG_FILTER_MY_PET) or CombatLog_Object_IsA(sourceFlags, COMBATLOG_FILTER_MINE)
  )

  local destMatch = destGUID == playerID or (
  CombatLog_Object_IsA(destFlags, COMBATLOG_FILTER_MY_PET) or CombatLog_Object_IsA(destFlags, COMBATLOG_FILTER_MINE)
  )
  if destMatch and kind == "damage" then return end

  if not unitMatch and not petMatch then return end

  if type(spellName) == "string" then
    spellName = spellName:gsub(" Off-Hand",""):gsub(" Overload","")
  end

  local seconds = tostring(math.floor(time))
  local finalSuff = seconds

  if unitMatch then
    if kind == "aura" then
      if destGUID and destGUID ~= playerID then
        unique:Set({ "aura", spellName or sourceName, destGUID }, seconds)
      end
      return
    elseif kind == "spell" then
      unique:Set({ "spell", spellName }, seconds)
      return
    end
  elseif kind =="aura" or kind == "spell" then
    return
  end
  local useAmount = amount
  if amount and overkill then useAmount = useAmount - overkill end
  for trackerName,trackerFrame in pairs(frames) do
    local tracker = self.core[trackerName]
    local trackerKey = K{ trackerName }
    if tracker and tracker.enabled and (not kind or kind == tracker.trackerType) then
      local duration = tracker.duration
      for meterType,name, meter in self:GetAll(tracker, "meter", 3) do
        local effectiveName = meter.name or name
        local match,id = self:Matches(meterType, effectiveName, meter, sourceName, petMatch, spellName, spellID)
        if match then
          local suff = ""
          local spell, aura, recent
          spell = unique:Get{ "spell", name }
          if destGUID then aura = unique:Get{ "aura", name, destGUID } end
          recent = unique:Get{ "recent", name }
          if spell then
            suff = suff .. spell
            finalSuff = ""
          elseif aura then
            suff = suff .. aura
            finalSuff = ""
          elseif recent and seconds - recent[2] < 2 then
            suff = suff .. recent[1]
            finalSuff = ""
          elseif recent then
            recent = { seconds }
          end
          if not recent then recent = { seconds } end
          recent[2] = seconds

          unique:Set({ "recent", name }, recent)
          local display = meter.display
          if amount and amount > 0 then
            local alertName = name
            if meterType == "spell" and display ~= "aggregate" then
              alertName = alertName .. suff .. finalSuff
              finalSuff = ""
            end
            if display == "individual" then
              alertName = alertName .. finalSuff .. unique:Incr{ "time", time }
              finalSuff = ""
            end
            self:BuildAlert(trackerName, meterType, match, name, id, alertName, useAmount, crit, time, overkill)
          end
        end
      end
    end
  end
end
